#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include "config.h"

#ifdef __CYGWIN32__
#include <Windows.h>
#endif

int comm(void);
void init_comm(char* SerialDevice,int BaudRate);
void end_comm(void);
int start_comm(int length);
int select_comm(void);
int select_stdin(void);
int write_comm(void* buf,int len);
int read_comm(void* buf,int len);
int truncate_baudrate(int baud);
void write_verify(unsigned char ch);
void upload_bootstrap(FILE* fp);
void upload_program(FILE* fp);
void debug_io(void);
void print_usage();

#define DEBUG_MAX_ROW 16
#define ERROR_RETURN end_comm();exit(1)
#define PING_WAIT_LOOP_CNT 2000
#define PING_TRY_LOOP_CNT 10

#ifdef __CYGWIN32__
HANDLE commfd=NULL;
#else
int commfd=0;
#endif

#define DWORD long

int BootstrapBaudrate,ProgramBaudrate;

int main(int argc,char* argv[])
{
    int c;
    char *bfn,*pfn;
    FILE* bfp;
    FILE* pfp;
    int DoNotWrite=0,DoDebug=0;
    char* SerialDevice=DEFAULT_DEVICE;
    extern int optind;
    extern char* optarg;

    /* sure to close COMM at exit */
    atexit(end_comm);
    /* set default values */
    ProgramBaudrate=BOOTSTRAP_BAUDRATE_MAX;
    BootstrapBaudrate=0;
    bfn=DEFAULT_BOOTSTRAP_FILE;

    /* analyze options */
    while((c=getopt(argc,argv,"hdnr:b:B:D:"))>=0){
	switch(c){
	case 'h':
	    print_usage();
	    break;
	case 'n':
	    DoNotWrite=1;
	    break;
	case 'd':
	    DoDebug=1;
	    break;
	case 'b':
	    ProgramBaudrate=atoi(optarg);
	    break;
	case 'B':
	    BootstrapBaudrate=atoi(optarg);
	    break;
	case 'r':
	    bfn=optarg;
	    break;
	case 'D':
	    SerialDevice=optarg;
	    break;
	}
    }
    /* Baudrate settings */
    if(BootstrapBaudrate==0){	/* if BootstrapBaudrate is not specified */
	BootstrapBaudrate=ProgramBaudrate;
    }
    if(BootstrapBaudrate>BOOTSTRAP_BAUDRATE_MAX){
	BootstrapBaudrate=BOOTSTRAP_BAUDRATE_MAX;
    }
    ProgramBaudrate=truncate_baudrate(ProgramBaudrate);
    BootstrapBaudrate=truncate_baudrate(BootstrapBaudrate);

    /* File name settings */
    bfp=fopen(bfn,"rb");
    if(bfp==NULL){
	if(bfn==DEFAULT_BOOTSTRAP_FILE){
	    fprintf(stderr,PROGNAME": Default bootstrap program is not\
 properly installed.\n Place it at "DEFAULT_BOOTSTRAP_FILE"\n");
	    ERROR_RETURN;
	}
	perror(PROGNAME);
	ERROR_RETURN;
    }
    pfn=(optind<argc)?argv[optind]:0;
    if(optind>=argc){	/* if program file is not specified */
	pfp=0;
    }
    else{
	pfp=fopen(argv[optind],"rb");
	if(pfp==NULL){
	    perror(PROGNAME);
	    ERROR_RETURN;
	}
    }

    /* main sequence */
    init_comm(SerialDevice,BootstrapBaudrate);
    if(!DoNotWrite){
	upload_bootstrap(bfp);
	if(pfp!=NULL){
	    upload_program(pfp);
	}
    }
    else{
	c=0;
	write_comm(&c,1);
    }
    if(DoDebug){
	debug_io();
    }
    end_comm();
}
void print_usage()
{
    fprintf(stderr,"Usage:\n");
    fprintf(stderr,PROGNAME" [-r FILE] [-b BAUDRATE] [-B BAUDRATE] [FILE]\n");
    fprintf(stderr,"-h : Print this help.\n");
    fprintf(stderr,"-r FILE\t: specify bootstrap program. Default %s\n",
	    DEFAULT_BOOTSTRAP_FILE);
    fprintf(stderr,"-b BPS\t: specify baudrate at program transfer.\
 Default %d\n",BOOTSTRAP_BAUDRATE_MAX);
    fprintf(stderr,"-B BPS\t: specify baudrate at bootstrap.\
 Default is a same as a -b option\n");
    fprintf(stderr,"-D Device\t: specify device name of the serial line.\
 Default "DEFAULT_DEVICE"\n");
    fprintf(stderr,"-n : Do not execute write procedure\n");
    fprintf(stderr,"-d : Connect stdin/stdout to serial port after writing\n");
    fprintf(stderr,"FILE : specify program file to be programmed\n");
    fprintf(stderr," BPS must be 2400 or higher,\n");
    fprintf(stderr,"  and may be truncated to an appropriate value\n");
    ERROR_RETURN;
}
int truncate_baudrate(int baudrate)
{
#ifdef __CYGWIN32__
    return baudrate;
#else
    if(baudrate<4800){
	return B2400;
    }
    else if(baudrate<9600){
	return B4800;
    }
    else if(baudrate<19200){
	return B9600;
    }
    else if(baudrate<38400){
	return B19200;
    }
    else if(baudrate<57600){
	return B38400;
    }
    else if(baudrate<115200){
	return B57600;
    }
    else if(baudrate<230400){
	return B115200;
    }
    else{
	return B230400;
    }
#endif
}

void upload_bootstrap(FILE* fp)
{
    int i,j,len;
    long length;
    unsigned char dat,rdat;
    fd_set fds;
    struct timeval tm;
    DWORD status;

    /* get file size */
    if(fseek(fp,0,SEEK_END)){
	perror(PROGNAME);
	ERROR_RETURN;
    }
    length=ftell(fp);
    if(length>MAX_BOOTSTRAP_PROGRAM_LENGTH){
	fprintf(stderr,PROGNAME": Bootstrap program exceeds %d byte\n",
		MAX_BOOTSTRAP_PROGRAM_LENGTH);
	ERROR_RETURN;
    }
    len=(int)length;
    rewind(fp);
    
    sleep(1);
    fprintf(stderr,"Initiating connection... ");
    FD_ZERO(&fds);
    /* initiate communication protocol */
    for(j=0;j<PING_TRY_LOOP_CNT&&!select_comm();j++){
	for(dat=0,i=0;i<PING_WAIT_LOOP_CNT;i++){
	    write_comm(&dat,1);
	}
    }
    if(j>=PING_TRY_LOOP_CNT){
	fprintf(stderr,PROGNAME": No answer.\n");
	ERROR_RETURN;
    }
    dat=0;
    read_comm(&dat,1);
    if(dat!=0){
	fprintf(stderr,PROGNAME": Transmission initiation error. %02X\n",dat);
    }
    dat=0x55;
    write_comm(&dat,1);
    dat=0;
    read_comm(&dat,1);
    if(dat!=0xaa){
	fprintf(stderr,PROGNAME": Transmission start error. %02X\n",dat);
	ERROR_RETURN;
    }
    fprintf(stderr,"Done.\n");
    write_verify(length/256);
    write_verify(length%256);
    for(i=0;i<len;i++){
	write_verify(fgetc(fp));
    }
    dat=0;
    read_comm(&dat,1);
    if(dat!=0xaa){
	fprintf(stderr,PROGNAME": Transmission exit error. %02X\n",dat);
	ERROR_RETURN;
    }
    fprintf(stderr,"The bootstrap program has been sent\n");
}
void upload_program(FILE* fp)
{
    long length,i;
    unsigned char dat;
    /* get file size */
    if(fseek(fp,0,SEEK_END)){
	perror(PROGNAME);
	ERROR_RETURN;
    }
    length=ftell(fp);
    if(length>MAX_ROM_PROGRAM_LENGTH){
	fprintf(stderr,PROGNAME": Bootstrap program exceeds %d byte\n",
		MAX_ROM_PROGRAM_LENGTH);
	ERROR_RETURN;
    }
    rewind(fp);

    /* upload length (4 byte) */
    dat=length>>24;
    write_comm(&dat,1);
    dat=(length>>16)&0xff;
    write_comm(&dat,1);
    dat=(length>>8)&0xff;
    write_comm(&dat,1);
    dat=length&0xff;
    write_comm(&dat,1);

    /* upload data */
    for(i=0;i<length;i++){
	dat=fgetc(fp);
	write_comm(&dat,1);
	dat=1;
	read_comm(&dat,1);
	if(dat!=0){
	    /* Write is not succeeded */
	    fprintf(stderr,PROGNAME": Write error. %02X\n",dat);
	    ERROR_RETURN;
	}
	/* print dot at every 256 bytes */
	if(i%256==0){
	    fprintf(stderr," %7d bytes %2d%% done\r",i,(i*100)/length);
/*
	    fprintf(stderr,".");
*/
	}
    }
    fprintf(stderr,"\n");
}
#ifdef __CYGWIN32__
void debug_io(void)
{
    fprintf(stderr,"Hit return key to end debug mode\n");
    getchar();
}
#else
void debug_io(void)
{
    unsigned char ch;
    fd_set fds;
    int row=0;

    FD_ZERO(&fds);
    fprintf(stderr,"debug mode\n");
    while(1){
	if(select_comm()){
	    read_comm(&ch,1);
	    putchar(ch);
	    fflush(stdout);
	}
	FD_SET(fileno(stdin),&fds);
	select(stdin+1,&fds,0,0,NULL);
	if(FD_ISSET(fileno(stdin),&fds)){
	    ch=getchar();
	    write_comm(&ch,1);
	}
    }
}
#endif

void write_verify(unsigned char ch)
{
    unsigned char rch;
    write_comm(&ch,1);
    rch=0;
    read_comm(&rch,1);
    if(ch!=rch){
	fprintf(stderr,PROGNAME": Transmission error. %02X!=%02X\n",ch,rch);
	ERROR_RETURN;
    }
}

/*

   Console support

*/
int select_stdin(void)
{
    fd_set fds;
    struct timeval tm;
    tm.tv_sec=0;
    tm.tv_usec=0;
    FD_ZERO(&fds);
    FD_SET(fileno(stdin),&fds);
    select(fileno(stdin)+1,&fds,0,0,&tm);
    return FD_ISSET(fileno(stdin),&fds);
}

/*
   
   Serial support routines
   
*/

#ifdef __CYGWIN32__

void init_comm(char* SerialDevice,int baudrate)
{
    DCB commDCB;

    commfd=CreateFile(SerialDevice,GENERIC_READ|GENERIC_WRITE,0,NULL,
		      OPEN_EXISTING,0,NULL);
    if(commfd==INVALID_HANDLE_VALUE){
	perror("init_comm");
	ERROR_RETURN;
    }
    EscapeCommFunction(commfd,CLRDTR);
    GetCommState(commfd,&commDCB);
    commDCB.BaudRate=baudrate;
    commDCB.fParity=FALSE;
    commDCB.fOutX=FALSE;
    commDCB.fInX=FALSE;
    commDCB.Parity=NOPARITY;
    commDCB.StopBits=ONESTOPBIT;
    SetCommState(commfd,&commDCB);
    EscapeCommFunction(commfd,SETDTR);
}

void end_comm()
{
    if(commfd!=NULL){
	EscapeCommFunction(commfd,CLRDTR);
	CloseHandle(commfd);
    }
}
int write_comm(void* buf,int len)
{
    DWORD dwWritten;

    if(!WriteFile(commfd,buf,len,&dwWritten,NULL)){
	return -1;
    }
    return dwWritten;
}

int read_comm(void* buf,int len)
{
    DWORD dwRead;
    if(!ReadFile(commfd,buf,len,&dwRead,NULL)){
	return 0;
    }
    else{
	return dwRead;
    }
}

int select_comm(void)
{
    DWORD dwCommEvent;
    int ans=0;
    OVERLAPPED osStatus={0};
    COMSTAT comStat;

    ClearCommError(commfd,&dwCommEvent,&comStat);
    return comStat.cbInQue!=0;
}

#else /* #if __CYGWIN__32 */
void init_comm(char* SerialDevice,int baudrate)
{
    struct termios termios;
    int pgrp;

    commfd=open(SerialDevice,O_RDWR);
    if(commfd<0){
	perror("init_comm");
    }
    pgrp=tcgetpgrp(commfd);
    if(pgrp>=0){
	fprintf(stderr,PROGNAME": Another program uses serial device.");
	ERROR_RETURN;
    }
    tcsetpgrp(commfd,getpgrp());
    tcgetattr(commfd,&termios);
    cfsetispeed(&termios,baudrate);
    cfsetospeed(&termios,baudrate);
    termios.c_cflag|=HUPCL|CLOCAL;
    termios.c_cflag&=~(PARENB|CSTOPB);
    cfmakeraw(&termios);
    tcsetattr(commfd,TCSANOW,&termios);
}

void end_comm()
{
    struct termios termios;

    if(commfd>0){
	tcgetattr(commfd,&termios);
	termios.c_oflag&=~CREAD;
	tcsetattr(commfd,TCSANOW,&termios);
	close(commfd);
    }
}

int write_comm(void* buf,int len)
{
    int ans;
    ans=write(commfd,buf,len);
    tcdrain(commfd);
    return ans;
}
int read_comm(void* buf,int len)
{
    return read(commfd,buf,len);
}

int select_comm(void)
{
    fd_set fds;
    struct timeval tm;
    tm.tv_sec=0;
    tm.tv_usec=0;
    FD_ZERO(&fds);
    FD_SET(commfd,&fds);
    select(commfd+1,&fds,0,0,&tm);
    return FD_ISSET(commfd,&fds);
}

#endif

